

#include "cuCache.h"
#include "cumath.h"



namespace TNet {

  CuCache::
  CuCache()
    : mState(EMPTY), mIntakePos(0), mExhaustPos(0), mDiscarded(0), 
      mRandomized(false), mTrace(0)
  { }

  CuCache::
  ~CuCache()
  { }

  void
  CuCache::
  Init(size_t cachesize, size_t bunchsize)
  {
    if((cachesize % bunchsize) != 0) {
      KALDI_ERR << "Non divisible cachesize(" << cachesize << ")" 
                << " by bunchsize(" << bunchsize <<")";
    }
    
    mCachesize = cachesize;
    mBunchsize = bunchsize;

    mState = EMPTY;

    mIntakePos = 0;
    mExhaustPos = 0;

    mRandomized = false;

  }

  void 
  CuCache::
  AddData(const CuMatrix<BaseFloat>& rFeatures, const Matrix<BaseFloat>& rDesired)
  {
    assert(rFeatures.Rows() == rDesired.Rows());

    //lazy buffers allocation
    if(mFeatures.Rows() != mCachesize) {
      mFeatures.Init(mCachesize,rFeatures.Cols());
      mDesired.Init(mCachesize,rDesired.Cols());
    }

    //warn if segment longer than half-cache
    if(rFeatures.Rows() > mCachesize/2) {
      KALDI_WARN << "Too long segment and small feature cache! "
                 << " cachesize: " << mCachesize
                 << " segmentsize: " << rFeatures.Rows();
    }

    //change state
    if(mState == EMPTY) { 
      if(mTrace&3) KALDI_COUT << "/";
      mState = INTAKE; mIntakePos = 0;
     
      //check for leftover from previous segment 
      int leftover = mFeaturesLeftover.Rows();
      //check if leftover is not bigger than cachesize
      if(leftover > mCachesize) {
        KALDI_WARN << "Too small feature cache: " << mCachesize
                   << ", truncating: "
                   << leftover - mCachesize << " frames from previous segment leftover";
        leftover = mCachesize;
      }
      //prefill cache with leftover
      if(leftover > 0) {
        mFeatures.CopyRows(leftover,0,mFeaturesLeftover,0);
        mDesired.RowRange(0,leftover).Copy(mDesiredLeftover.RowRange(0,leftover));
        mFeaturesLeftover.Destroy();
        mDesiredLeftover.Destroy();
        mIntakePos += leftover;
      } 
    }

    assert(mState == INTAKE);
    assert(rFeatures.Rows() == rDesired.Rows());
    if(mTrace&2) KALDI_COUT << "F";

    int cache_space = mCachesize - mIntakePos;
    int feature_length = rFeatures.Rows();
    int fill_rows = (cache_space<feature_length)? cache_space : feature_length;
    int leftover = feature_length - fill_rows;

    assert(cache_space > 0);

    //copy the data to cache
    mFeatures.CopyRows(fill_rows,0,rFeatures,mIntakePos);
    mDesired.RowRange(mIntakePos,fill_rows).Copy(rDesired.RowRange(0,fill_rows));

    //copy leftovers
    if(leftover > 0) {
      mFeaturesLeftover.Init(leftover,mFeatures.Cols());
      mDesiredLeftover.Init(leftover,mDesired.Cols());
      mFeaturesLeftover.CopyRows(leftover,fill_rows,rFeatures,0);
      mDesiredLeftover.RowRange(0,leftover).Copy(rDesired.RowRange(fill_rows,leftover));
    }
 
    //update cursor
    mIntakePos += fill_rows;
    
    //change state
    if(mIntakePos == mCachesize) { 
      if(mTrace&3) KALDI_COUT << "\\";
      mState = FULL;
    }
  }



  void
  CuCache::
  Randomize()
  {
    assert(mState == FULL || mState == INTAKE);

    if(mTrace&3) KALDI_COUT << "R";

    //lazy initialization of hte output buffers
    mFeaturesRandom.Init(mCachesize,mFeatures.Cols());
    mDesiredRandom.Init(mCachesize,mDesired.Cols());

    //generate random series of integers
    Vector<int> randmask(mIntakePos);
    for(unsigned int i=0; i<mIntakePos; i++) {
      randmask[i]=i;
    }
    int* ptr = randmask.pData();
    std::random_shuffle(ptr, ptr+mIntakePos, GenerateRandom);

    CuVector<int> cu_randmask;
    cu_randmask.CopyFrom(randmask);

    //randomize
    CuMath<BaseFloat>::Randomize(mFeaturesRandom,mFeatures,cu_randmask);
    for (int i=0; i<randmask.Dim(); i++) {
      mDesiredRandom[i].Copy(mDesired[randmask[i]]);
    }

    mRandomized = true;
  }

  void
  CuCache::
  GetBunch(CuMatrix<BaseFloat>& rFeatures, Matrix<BaseFloat>& rDesired)
  {
    if(mState == EMPTY) {
      KALDI_ERR << "GetBunch on empty cache!!!";
    }

    //change state if full...
    if(mState == FULL) { 
      if(mTrace&3) KALDI_COUT << "\\";
      mState = EXHAUST; mExhaustPos = 0; 
    }

    //final cache is not completely filled
    if(mState == INTAKE) //&& mpFeatures->EndOfList()
    { 
      if(mTrace&3) KALDI_COUT << "\\-LAST\n";
      mState = EXHAUST; mExhaustPos = 0; 
    } 

    assert(mState == EXHAUST);

    //init the output
    rFeatures.Init(mBunchsize,mFeatures.Cols());
    rDesired.Init(mBunchsize,mDesired.Cols());

    //copy the output
    if(mRandomized) {
      rFeatures.CopyRows(mBunchsize,mExhaustPos,mFeaturesRandom,0);
      rDesired.Copy(mDesiredRandom.RowRange(mExhaustPos,mBunchsize));
    } else {
      rFeatures.CopyRows(mBunchsize,mExhaustPos,mFeatures,0);
      rDesired.Copy(mDesired.RowRange(mExhaustPos,mBunchsize));
    }

    //update cursor
    mExhaustPos += mBunchsize;

    //change state to EMPTY
    if(mExhaustPos > mIntakePos-mBunchsize) {
      //we don't have more complete bunches...
      mDiscarded += mIntakePos - mExhaustPos;

      mState = EMPTY;
    }
  }

}
