#ifndef _LABELS_H_
#define _LABELS_H_


#include "Matrix.h"
#include "MlfStream.h"
#include "Features.h"

#include <map>
#include <iostream>

namespace TNet {

  /**
   * Desired matrix generation object,
   * supports background-reading and caching, however can be 
   * used in foreground as well by GenDesiredMatrix()
   */
  class LabelRepository 
  {
    typedef std::map<std::string,size_t> TagToIdMap;

    public:
      LabelRepository()
        : _mpLabelStream(NULL), mpLabelStream(NULL), mpLabelDir(NULL), mpLabelExt(NULL), mGenDesiredMatrixTime(0), mIndexTime(0), mTrace(0), mIsReady(false) 
      { }

      ~LabelRepository()
      { 
        if(mTrace&4) {
          KALDI_COUT << "[LabelRepository -- indexing:" << mIndexTime << "s"
                       " genDesiredMatrix:" << mGenDesiredMatrixTime << "s]" << std::endl;
        }
        delete mpLabelStream;
        delete _mpLabelStream;
      }

      /// Initialize the LabelRepository      
      void Init(const char* pLabelMlfFile, const char* pOutputLabelMapFile, const char* pLabelDir, const char* pLabelExt);

      /// Check if LabelRepository is iniliazied
      bool IsReady()
      { return mIsReady; }

      /// Set trace level
      void Trace(int trace)
      { mTrace = trace; }

      /// Get desired matrix from labels
      bool GenDesiredMatrix(BfMatrix& rDesired, size_t nFrames, size_t sourceRate, const char* pFeatureLogical);

    private:
      /// Prepare the state-label to state-id map
      void ReadOutputLabelMap(const char* file);
      
    private:
      // Streams and state-map
      std::ifstream* _mpLabelStream; ///< Helper stream for Label stream
      IMlfStream* mpLabelStream;     ///< Label stream
      std::istringstream mGenDesiredMatrixStream; ///< Label file parsing stream
     
      const char* mpLabelDir;  ///< Label dir in MLF 
      const char* mpLabelExt;  ///< Label ext in MLF
      char mpLabelFile[4096];  ///< Buffer for filenames in MLF
      
      TagToIdMap mLabelMap; ///< Map of state tags to net output indices

      double mGenDesiredMatrixTime;
      float  mIndexTime;

      int mTrace;
      bool mIsReady;
  };

  /**
   * Try to get matrices with same row-number,
   * truncate rows up to some tolerance.
   */
  inline
  void UnifyMatrixRowNumber(Matrix<BaseFloat> *mat1, Matrix<BaseFloat> *mat2, int tolerance = 5, int mat1_ext = 0) {
    assert(mat1 != NULL && mat2 != NULL);
    
    //see how big the difference is, 
    int row_difference = (mat1->Rows()-mat1_ext) - mat2->Rows();
 
    //in some cases don't do anything
    if(row_difference == 0) return; //no difference! it's okay!
    if(abs(row_difference) > tolerance) return; //too big difference, bad luck!

    //let's truncate the longer one
    if(row_difference > 0) {
      //mat1 is longer, truncate it!
      for(int i=0; i<row_difference; i++) {
        mat1->RemoveRow(mat1->Rows()-1); //remove last row,
      }
    } else {
      //mat2 is longer, truncate it!
      for(int i=0; i<abs(row_difference); i++) {
        mat2->RemoveRow(mat2->Rows()-1); //remove last row,
      }
    }

    assert(mat1->Rows()-mat1_ext == mat2->Rows());
  }

}//namespace

#endif
