

#include "L2MeanLinearity.h"


namespace TNet {


void
L2MeanLinearity::
PropagateFnc(const Matrix<BaseFloat>& X, Matrix<BaseFloat>& Y)
{
  //y = b + x.A

  //precopy bias
  size_t rows = X.Rows();
  for(size_t i=0; i<rows; i++) {
    Y[i].Copy(*mpBias);
  }

  //multiply matrix by matrix with mLinearity
  Y.BlasGemm(1.0f, X, NO_TRANS, *mpLinearity, NO_TRANS, 1.0f);
}


void
L2MeanLinearity::
BackpropagateFnc(const Matrix<BaseFloat>& X, Matrix<BaseFloat>& Y)
{
  // e' = e.A^T
  Y.Zero();
  Y.BlasGemm(1.0f, X, NO_TRANS, *mpLinearity, TRANS, 0.0f);
}



void
L2MeanLinearity::
ReadFromStream(std::istream& rIn)
{
  //matrix is stored transposed as SNet does
  Matrix<BaseFloat> transpose;
  rIn >> transpose;
  mLinearity = Matrix<BaseFloat>(transpose, TRANS);
  //biases stored normally
  rIn >> mBias;

  //read weights to regularize to or copy current weights?
  //see what is in the file
  while(isascii(rIn.peek()) && isspace(rIn.peek())) rIn.get(); // eat up space.
  if(rIn.peek() == 'm') {
    rIn >> transpose;
    mLinearityL2Mean = Matrix<BaseFloat>(transpose, TRANS);
  } else {
    mLinearityL2Mean.Copy(mLinearity);
  }
  while(isascii(rIn.peek()) && isspace(rIn.peek())) rIn.get(); // eat up space.
  if(rIn.peek() == 'v') {
    rIn >> mBiasL2Mean;
  } else {
    mBiasL2Mean.Copy(mBias);
  }
}

 
void
L2MeanLinearity::
WriteToStream(std::ostream& rOut)
{
  //matrix is stored transposed as SNet does
  Matrix<BaseFloat> transpose(mLinearity, TRANS);
  rOut << transpose;
  //biases stored normally
  rOut << mBias;
  rOut << std::endl;
  
  //also store the L2 regularization means
  transpose = Matrix<BaseFloat>(mLinearityL2Mean,TRANS);
  rOut << transpose;
  rOut << mBiasL2Mean;
  rOut << std::endl;
}


void
L2MeanLinearity::
Gradient()
{
  //calculate gradient of weight matrix
  mLinearityCorrection.Zero();
  mLinearityCorrection.BlasGemm(1.0f, GetInput(), TRANS, 
                                GetErrorInput(), NO_TRANS, 
                                0.0f);

  //calculate gradient of bias
  mBiasCorrection.Set(0.0f);
  size_t rows = GetInput().Rows();
  for(size_t i=0; i<rows; i++) {
    mBiasCorrection.Add(GetErrorInput()[i]);
  }

  /* 
  //perform update
  mLinearity.AddScaled(-mLearningRate, mLinearityCorrection);
  mBias.AddScaled(-mLearningRate, mBiasCorrection);
  */
}


void 
L2MeanLinearity::
AccuGradient(const UpdatableComponent& src, int thr, int thrN) {
  //cast the argument
  const L2MeanLinearity& src_comp = dynamic_cast<const L2MeanLinearity&>(src);

  //allocate accumulators when needed
  if(mLinearityCorrectionAccu.MSize() == 0) {
    mLinearityCorrectionAccu.Init(mLinearity.Rows(),mLinearity.Cols());
  }
  if(mBiasCorrectionAccu.MSize() == 0) {
    mBiasCorrectionAccu.Init(mBias.Dim());
  }

  //need to find out which rows to sum...
  int div = mLinearityCorrection.Rows() / thrN;
  int mod = mLinearityCorrection.Rows() % thrN;

  int origin = thr * div + ((mod > thr)? thr : mod);
  int rows = div + ((mod > thr)? 1 : 0);

  //we may have more threads than lines in a matrix,
  //so some threads will not do anything
  if(rows == 0) return;

  //create the matrix windows
  const SubMatrix<BaseFloat> src_mat (
    src_comp.mLinearityCorrection, 
    origin, rows, 
    0, mLinearityCorrection.Cols()
  );
  SubMatrix<double> tgt_mat (
    mLinearityCorrectionAccu, 
    origin, rows, 
    0, mLinearityCorrection.Cols()
  );
  //sum the rows
  Add(tgt_mat,src_mat);

  //first thread will always sum the bias correction
  if(thr == 0) {
    Add(mBiasCorrectionAccu,src_comp.mBiasCorrection);
  }

}


void
L2MeanLinearity::
Update(int thr, int thrN)
{
  //need to find out which rows to sum...
  int div = mLinearity.Rows() / thrN;
  int mod = mLinearity.Rows() % thrN;

  int origin = thr * div + ((mod > thr)? thr : mod);
  int rows = div + ((mod > thr)? 1 : 0);

  //we may have more threads than lines in a matrix,
  //so some threads will not do anything
  if(rows == 0) return;

  //get the matrix windows
  SubMatrix<double> src_mat (
    mLinearityCorrectionAccu, 
    origin, rows, 
    0, mLinearityCorrection.Cols()
  );
  SubMatrix<BaseFloat> tgt_mat (
    mLinearity, 
    origin, rows, 
    0, mLinearityCorrection.Cols()
  );
  SubMatrix<BaseFloat> l2mean_mat (
    mLinearityL2Mean, 
    origin, rows, 
    0, mLinearityCorrection.Cols()
  );



  //update weights
  AddScaled(tgt_mat, src_mat, -mLearningRate);

  //perform L2 regularization to the original values, means of prior N(mi,l*I),
  //this is adaptation mode
  BaseFloat L2_decay = -mLearningRate * mWeightcost * mBunchsize;
  if(L2_decay != 0.0) {
    Matrix<BaseFloat> l2_aux(tgt_mat);
    l2_aux.AddScaled(-1.0,l2mean_mat);
    tgt_mat.AddScaled(L2_decay, l2_aux);
  }

  //first thread always update bias
  if(thr == 0) {
    AddScaled(mBias, mBiasCorrectionAccu, -mLearningRate);
    //regularize also the biases (unlike the <BiasedLinearity>, where is not L2)
    Vector<BaseFloat> l2_bias_aux(mBias);
    l2_bias_aux.AddScaled(-1.0,mBiasL2Mean);
    mBias.AddScaled(L2_decay,l2_bias_aux);
  }

  //reset the accumulators
  src_mat.Zero();
  if(thr == 0) {
    mBiasCorrectionAccu.Zero();
  }

}

} //namespace
