#ifndef _L2MEAN_LINEARITY_H_
#define _L2MEAN_LINEARITY_H_


#include "Component.h"

#include "Matrix.h"
#include "Vector.h"


namespace TNet {

class L2MeanLinearity : public UpdatableComponent
{
 public:

  L2MeanLinearity(size_t nInputs, size_t nOutputs, Component *pPred);
  ~L2MeanLinearity() { } 
  
  ComponentType GetType() const
  { return L2MEAN_LINEARITY; }

  const char* GetName() const
  { return "<L2MeanLinearity>"; }

  Component* Clone() const;

  void PropagateFnc(const Matrix<BaseFloat>& X, Matrix<BaseFloat>& Y);
  void BackpropagateFnc(const Matrix<BaseFloat>& X, Matrix<BaseFloat>& Y);

  void ReadFromStream(std::istream& rIn);
  void WriteToStream(std::ostream& rOut);

  /// calculate gradient
  void Gradient();
  /// accumulate gradient from other components
  void AccuGradient(const UpdatableComponent& src, int thr, int thrN);  
  /// update weights, reset the accumulator
  void Update(int thr, int thrN);

 protected:
  Matrix<BaseFloat> mLinearity;  ///< Matrix with neuron weights
  Vector<BaseFloat> mBias;       ///< Vector with biases

  const Matrix<BaseFloat>* mpLinearity;
  const Vector<BaseFloat>* mpBias;

  Matrix<BaseFloat> mLinearityCorrection; ///< Matrix for linearity updates
  Vector<BaseFloat> mBiasCorrection;      ///< Vector for bias updates

  Matrix<double> mLinearityCorrectionAccu; ///< Matrix for summing linearity updates
  Vector<double> mBiasCorrectionAccu;      ///< Vector for summing bias updates

  Matrix<BaseFloat> mLinearityL2Mean;  ///< Matrix with neuron weights
  Vector<BaseFloat> mBiasL2Mean;       ///< Vector with biases
};




////////////////////////////////////////////////////////////////////////////
// INLINE FUNCTIONS 
// L2MeanLinearity::
inline 
L2MeanLinearity::
L2MeanLinearity(size_t nInputs, size_t nOutputs, Component *pPred)
  : UpdatableComponent(nInputs, nOutputs, pPred), 
    mLinearity(), mBias(), //cloned instaces don't need this
    mpLinearity(&mLinearity), mpBias(&mBias), 
    mLinearityCorrection(nInputs,nOutputs), mBiasCorrection(nOutputs),
    mLinearityCorrectionAccu(), mBiasCorrectionAccu(), //cloned instances don't need this
    mLinearityL2Mean(nInputs,nOutputs), mBiasL2Mean(nOutputs)
{ }

inline
Component* 
L2MeanLinearity::
Clone() const
{
  L2MeanLinearity* ptr = new L2MeanLinearity(GetNInputs(), GetNOutputs(), NULL);
  ptr->mpLinearity = mpLinearity; //copy pointer from currently active weights
  ptr->mpBias = mpBias;           //...

  ptr->mLinearityL2Mean = mLinearityL2Mean;
  ptr->mBiasL2Mean = mBiasL2Mean;

  ptr->mLearningRate = mLearningRate;
  ptr->mMomentum = mMomentum;
  ptr->mWeightcost = mWeightcost;
  ptr->mBunchsize = mBunchsize;
  
  return ptr;
}



} //namespace



#endif
