#!/bin/bash

#threads were not specified, assume using CUDA
if [[ ! $THREADS ]]; then
  echo "CUDA disabled... <tanh> not implemented in TNet"
  #CUDA=1
fi
TNet=$TNET_ROOT/src/TNet${CUDA:+Cu}
echo
echo This script: $0
echo
echo TNet: $TNet

##############################
#check for obligatory parameters
echo NN_INIT: ${NN_INIT?$0: NN_INIT not specified}
#echo MLF_CV: ${MLF_CV?$0: MLF_CV not specified}
#echo MLF_TRAIN: ${MLF_TRAIN?$0: MLF_TRAIN not specified}
echo SCP_CV_LOCAL: ${SCP_CV_LOCAL?$0: SCP_CV_LOCAL not specified}
echo SCP_TRAIN_LOCAL: ${SCP_TRAIN_LOCAL?SCP_TRAIN_LOCAL not specified}
echo LEARNRATE: ${LEARNRATE?$0: LEARNRATE not specified}

##############################
#define implicit configuration
echo BUNCHSIZE: ${BUNCHSIZE:=512}
echo CACHESIZE: ${CACHESIZE:=32768}
echo RANDOMIZE: ${RANDOMIZE:=TRUE}
echo TRACE: ${TRACE:=5}
echo TNET_FLAGS: ${TNET_FLAGS:=-A -D -V --OBJECTIVE-FUNCTION=mse}

echo MAX_ITER: ${MAX_ITER:=20}
echo MIN_ITER: ${MIN_ITER:=1}
echo KEEP_LRATE_ITER: ${KEEP_LRATE_ITER:=0}
echo END_HALVING_INC: ${END_HALVING_INC:=0.001}
echo START_HALVING_INC: ${START_HALVING_INC:=0.1}
echo HALVING_FACTOR: ${HALVING_FACTOR:=0.5}

echo
echo "[[MSE REL. IMPROVEMENT SCHEDULING]]"
echo "START_HALVING_INC,END_HALVING_INC are relative improvements..."
echo


#runs the training commad, parses the mean-square-error
function run_tnet_parse_mse {
  local cmd=$1;
  echo %%%%%%
  echo $cmd
  echo %%%%%%
  local logfile=$(mktemp);
  $cmd | tee $logfile | sed 's|^|  |'
  #parse the logfile to get mean-square-error: 
  MSE=$(cat $logfile | grep 'Mse:' | tail -n 1 | sed 's|.*err\/frm:\([0-9\.]*\).*|\1|')
  rm $logfile
  if [[ $MSE == "" ]]; then
    echo "Error, No mean-square-error returned, terminating...";
    exit 1;
  fi
}


if [ ! -d weights ]; then
  mkdir weights
fi

nnet_base=weights/$(basename $NN_INIT .init)

echo "########################################################"
echo "# INITIAL CROSSVAL, $(date)"
echo "########################################################"

cmd="$TNet $TNET_FLAGS -T $TRACE \
 -H $NN_INIT \
 -S $SCP_CV_LOCAL \
 --BUNCHSIZE=$BUNCHSIZE \
 --CACHESIZE=$CACHESIZE \
 --RANDOMIZE=FALSE \
 --CROSSVALIDATE=TRUE \
 --STARTFRMEXT=$FRM_EXT \
 --ENDFRMEXT=$FRM_EXT \
 ${FEATURE_TRANSFORM:+--FEATURETRANSFORM=$FEATURE_TRANSFORM} \
 ${STK_CONF:+-C $STK_CONF} \
 ${THREADS:+--THREADS=$THREADS} \
 ${USE_GPU_ID:+--USE-GPU-ID=$USE_GPU_ID} \
 "

run_tnet_parse_mse "$cmd"
mse_best=$MSE
nnet_best=$NN_INIT
echo "Initial CV mean-square-error: $MSE"\


iter=0
do_halving=0
for iter in $(seq -w $MAX_ITER); do

  echo "########################################################"
  echo "# ITERATION:$iter, $(date)"
  echo "########################################################"
  nnet_next=${nnet_base}_iter$iter
  #run epoch
  cmd="$TNet $TNET_FLAGS -T $TRACE \
   -H $nnet_best \
   -S $SCP_TRAIN_LOCAL \
   --LEARNINGRATE=$LEARNRATE \
   ${LEARNRATEFACTORS:+--LEARNRATEFACTORS=$LEARNRATEFACTORS} \
   ${MOMENTUM:+--MOMENTUM=$MOMENTUM} \
   ${WEIGHTCOST:+--WEIGHTCOST=$WEIGHTCOST} \
   --BUNCHSIZE=$BUNCHSIZE \
   --CACHESIZE=$CACHESIZE \
   --RANDOMIZE=$RANDOMIZE \
   --TARGETMMF=$nnet_next \
   --STARTFRMEXT=$FRM_EXT \
   --ENDFRMEXT=$FRM_EXT \
   ${FEATURE_TRANSFORM:+--FEATURETRANSFORM=$FEATURE_TRANSFORM} \
   ${STK_CONF:+-C $STK_CONF} \
   ${THREADS:+--THREADS=$THREADS} \
   ${USE_GPU_ID:+--USE-GPU-ID=$USE_GPU_ID} \
   "

  run_tnet_parse_mse "$cmd"
  mse_train=$MSE
  echo "TR mean-square-error:  $MSE iter: $iter learnrate: $LEARNRATE"\
   ${MOMENTUM:+" momentum: $MOMENTUM"}\
   ${WEIGHTCOST:+" weightcost: $WEIGHTCOST"}

  echo "########################################################"
  echo "# CROSSVAL:$iter, $(date)"
  echo "########################################################"

  cmd="$TNet $TNET_FLAGS -T $TRACE \
   -H $nnet_next \
   -S $SCP_CV_LOCAL \
   --BUNCHSIZE=$BUNCHSIZE \
   --CACHESIZE=$CACHESIZE \
   --RANDOMIZE=FALSE \
   --CROSSVALIDATE=TRUE \
   --STARTFRMEXT=$FRM_EXT \
   --ENDFRMEXT=$FRM_EXT \
   ${FEATURE_TRANSFORM:+--FEATURETRANSFORM=$FEATURE_TRANSFORM} \
   ${STK_CONF:+-C $STK_CONF} \
   ${THREADS:+--THREADS=$THREADS} \
   ${USE_GPU_ID:+--USE-GPU-ID=$USE_GPU_ID} \
   "

  run_tnet_parse_mse "$cmd"
  mse_cv=$MSE
  echo "CV mean-square-error: $MSE iter: $iter learnrate: $LEARNRATE"\
   ${MOMENTUM:+" momentum: $MOMENTUM"}\
   ${WEIGHTCOST:+" weightcost: $WEIGHTCOST"}

  nnet_next_mse=${nnet_next}_lr$(printf '%.5g' $LEARNRATE)_tr$(printf '%.5g' $mse_train)_cv$(printf '%.5g' $mse_cv)
  mv $nnet_next $nnet_next_mse

  # always accept the weights when fixed LEARNRATE by keep_LEARNRATE_iter
  if [[ ${iter##0} -lt $KEEP_LRATE_ITER ]]; then
    echo 'accepting weights (keep_LEARNRATE_iter)'
    nnet_best=$nnet_next_mse
    mse_prev=$mse_best
    mse_best=$mse_cv
    continue
  fi

  #revert the parameters or keep them
  if [[ 1 == $(awk 'BEGIN{print('$mse_cv'<'$mse_best')}') ]]; then
    echo 'accepting weights (rel. improvement:'$(awk "BEGIN{print 1.0-$mse_cv/$mse_best;}")')';
    nnet_best=$nnet_next_mse
    mse_prev=$mse_best
    mse_best=$mse_cv
  else
    echo 'reverting the weights' " $mse_cv < $mse_best"
    mse_prev=$mse_best
    mv $nnet_next_mse ${nnet_next_mse}_rejected
  fi

  # end training if halving already and not improving much
  if [[ $do_halving == 1 && 1 == $(awk 'BEGIN{print(1.0-'$mse_best'/'$mse_prev'<'$END_HALVING_INC')}') && ${iter##0} -gt $MIN_ITER ]]; then
    break;
  fi


  # start halving when not improving much
  if [[ 1 == $(awk 'BEGIN{print(1.0-'$mse_cv'/'$mse_prev'<'$START_HALVING_INC')}') ]]; then
    do_halving=1
  fi

  if [[ $do_halving == 1 ]]; then
    LEARNRATE=$(awk 'BEGIN{print('$LEARNRATE'*'$HALVING_FACTOR')}')
    echo LEARNRATE=$LEARNRATE
  fi

done

#copy-out the best network
if [ $iter -gt 0 ]; then
  cp $nnet_best ${nnet_base}_final_iters${iter}_tr$(printf '%.5g' $mse_train)_cv$(printf '%.5g' $mse_best)
fi

echo "########################################################"
echo "# END OF TRAINING, $(date)"
echo "########################################################"
